#'
#' ANOVA statistic with privacy preserving
#'
#' ANOVA statistic that preserves the privacy of the data sources. The statistic is performed for at least two samples.
#'
#' @param params The <Parameters> that should be defined to compute the statistic.
#' @param method The method of anova that will be applied. Ex. 'b' is for one-way between subjects, 'r' is for repeated measures and 'm' for mixed.
#' @return A set of resulted components (summary table) after the ANOVA execution
#' @examples
#' Example for one-way anova between participants:
#' We want to apply an one-way anova between participants
#' on the dependend variable 'value_time_before' and on two groups regarding their gender (sex=M/F).
#' Firstly we should define the <Parameters>:
#' gr1 <- GroupStat( list( c("sex", "=", "M") ))
#' gr2 <- GroupStat( list( c("sex", "=", "F") ))
#' btwParam <- Parameters( list("value_time_before") , list(gr1, gr2))
#'
#' To run the one-way BP:
#' anova(btwParam) or anova(btwParam, "b")
#'
#' Example for one-way anova within participants or one-way repeated anova:
#' We want to apply an one-way anova within participants
#' on the dependend variable 'value' and on two conditions regarding the time (time=before/after).
#' Firstly we should define the <Parameters>:
#' gr1 <- GroupStat( list( c("time", "=", "before") ))
#' gr2 <- GroupStat( list( c("time", "=", "after") ))
#' repParam <- Parameters( list("value") , list(gr1, gr2))
#'
#' To run the one-way WP:
#' anova(repParam, "r")
#'
#' Example for the mixed anova:
#' We want to apply a mixed anova
#' on the dependend variable 'value', on two groups regarding their gender (sex=M/F)
#' and on two conditions regarding the time (time=before/after).
#' Firstly we should define the <Parameters>. Group 1 should contain the groups, while Group 2 the conditions:
#' gr1 <- GroupStat( list( c("sex", "=", "M"), c("sex", "=", "F") ))
#' gr2 <- GroupStat( list( c("time", "=", "before"), c("time", "=", "after") ))
#' mixParam <- Parameters( list("value") , list(gr1, gr2))
#'
#' To run the Mixed Anova:
#' anova(mixParam, "m")
#' @export
anova <- function(params, method='b')
{

  if(missing(params)) {
    stop("The argument 'params' should be filled in.")
  }

  if(class(params)[2] != "Parameters") {
    stop("The argument 'params' should be an instance of 'Parameters'.")
  }

  if(method == "between" || method == "b") {
    input <- .setInputOneWay(params)

    testObj <- .jnew("gr.demokritos.iit.radio.home.orchestration.Controller")
    result <- .jcall(testObj, "Ljava/util/HashMap;", "betweenAnovaParam", input)

    .convertOneWayResult(result)
  }
  else if(method == "repeated" || method == "r") {
    input <- .setInputOneWay(params)

    testObj <- .jnew("gr.demokritos.iit.radio.home.orchestration.Controller")
    result <- .jcall(testObj, "Ljava/util/HashMap;", "repeatedAnovaParam", input)

    .convertRepeatedResult(result)
  }
  else if(method == "mixed" || method == "m") {
    input <- .setInputOneWay(params)

    testObj <- .jnew("gr.demokritos.iit.radio.home.orchestration.Controller")
    result <- .jcall(testObj, "Ljava/util/HashMap;", "mixedAnovaParam", input)

    .convertMixedResult(result)
  }
  else {
    warning("Wrong input on 'method' parameter.")
  }

  }

#Set the inputs for the T-test
.setInputOneWay <- function(params) {

  # the input parameter contains the above specified objects
  input <- .jnew("gr.demokritos.iit.radio.home.entities.StatParams")

  #for select
  selectList <- .buildSelectOneWay(params)

  #for each group
  for (group in getGroups(params)) {
    # set the QfrMessage . Each one has the same select part.
    msg <- .jnew("gr.demokritos.iit.radio.home.qfr.QfrMessage")
    .jrcall(msg, "setSelect", selectList)
    msg <- .buildWhereOneWay(group, msg)
    print(msg)
    .jrcall(input, "addQfr", msg)
  }

  return(input)

}

#Builds the select clause of the query, as it is specified by the user
#@return: List<String> that defines the select clause
.buildSelectOneWay <- function(params) {
  cor <- getDependency(params)

  selectList <- .jnew("java.util.ArrayList")

  for(i in 1:length(cor)) {
    .jrcall(selectList, "add", as.character(cor[i]))
  }

  return(selectList)
}


#Takes all the basic information to convert the where clause of the query
#@return: List<WhereClause> that defines the where clause
.buildWhereOneWay <- function(group, msg) {

  #where <- .jnew("gr.demokritos.iit.radio.home.entities.qfr.WhereClause")

  triples <- getTriples(group)

  for (triple in triples) {
    vect <- .jnew("java.util.Vector")

    for(val in triple) {
      .jrcall(vect, "add", val)
    }
    .jcall(msg, "V","addTriple", vect)

  }

  return(msg)
}

.convertOneWayResult <- function(result) {
  keySet<-.jrcall(result,"keySet")
  an_iter<-.jrcall(keySet,"iterator")
  aList <- list()

  df <- c(.jrcall(result,"get",'df.g'),
          .jrcall(result,"get",'df.e') ,
          .jrcall(result,"get",'df.tot'))

  ssq <- c(.jrcall(result,"get",'SSB'),
           .jrcall(result,"get",'SSE'),
           .jrcall(result,"get",'SST') )

  ms <- c(.jrcall(result,"get",'MSB'),
          .jrcall(result,"get",'MSE'),
          ' ' )

  f <- c(.jrcall(result,"get",'F.value'),
         ' ',
         ' ' )

  p <- c(.jrcall(result,"get",'p.value'),
         ' ',
         ' ' )


  while(.jrcall(an_iter,"hasNext")) {
    key <- .jrcall(an_iter,"next");
    value <-.jrcall(result,"get", key)

    aList[[key]] <- value
  }

  mydf <- data.frame(Results = c('Groups','Errors', 'Total'),
                     Df = df,
                     Sum_Sq = ssq,
                     Mean_Sq = ms,
                     F_value = f,
                     P_value = p,
                     check.names = TRUE)
  grid.table(mydf)

  return(aList)
}


.convertRepeatedResult <- function(result) {
  keySet<-.jrcall(result,"keySet")
  an_iter<-.jrcall(keySet,"iterator")
  aList <- list()

  df <- c(.jrcall(result,"get",'df.cond'),
          .jrcall(result,"get",'df.subj') ,
          .jrcall(result,"get",'df.error'),
          .jrcall(result,"get",'df.tot') )

  ssq <- c(.jrcall(result,"get",'SScond'),
           .jrcall(result,"get",'SSsubj'),
           .jrcall(result,"get",'SSE'),
           .jrcall(result,"get",'SST') )

  ms <- c(.jrcall(result,"get",'MScond'),
          .jrcall(result,"get",'MSsubj'),
          .jrcall(result,"get",'MSE'),
          ' ')

  f <- c(.jrcall(result,"get",'F.cond'),
         .jrcall(result,"get",'F.subj'),
         ' ',
         ' ' )

  p <- c(.jrcall(result,"get",'p.value'),
         ' ',
         ' ',
         ' ' )


  while(.jrcall(an_iter,"hasNext")) {
    key <- .jrcall(an_iter,"next");
    value <-.jrcall(result,"get", key)

    aList[[key]] <- value
  }

  mydf <- data.frame(Results = c('Conditions','Subjects', 'Errors', 'Total'),
                     Df = df,
                     Sum_Sq = ssq,
                     Mean_Sq = ms,
                     F_value = f,
                     P_value = p,
                     check.names = TRUE)
  grid.table(mydf)

  return(aList)
}

.convertMixedResult <- function(result) {
  keySet<-.jrcall(result,"keySet")
  an_iter<-.jrcall(keySet,"iterator")
  aList <- list()

  df <- c(.jrcall(result,"get",'df.A'),
          .jrcall(result,"get",'df_S/A') ,
          .jrcall(result,"get",'df.U'),
          .jrcall(result,"get",'df_UxA'),
          .jrcall(result,"get",'df_UxS/A') )

  ssq <- c(.jrcall(result,"get",'SSA'),
           .jrcall(result,"get",'SS_S/A'),
           .jrcall(result,"get",'SSU'),
           .jrcall(result,"get",'SS_UxA'),
           .jrcall(result,"get",'SS_UxS/A') )

  ms <- c(.jrcall(result,"get",'MSA'),
          .jrcall(result,"get",'MS_S/A'),
          .jrcall(result,"get",'MSU'),
          .jrcall(result,"get",'MS_UxA'),
          .jrcall(result,"get",'MS_UxS/A') )

  f <- c(.jrcall(result,"get",'F_A'),
          ' ',
          .jrcall(result,"get",'F_U'),
          .jrcall(result,"get",'F_UxA'),
          ' ' )

  p <- c(.jrcall(result,"get",'p.value_A'),
         ' ',
         .jrcall(result,"get",'p.value_U'),
         .jrcall(result,"get",'p.value_UxA'),
         ' ' )


  while(.jrcall(an_iter,"hasNext")) {
    key <- .jrcall(an_iter,"next");
    value <-.jrcall(result,"get", key)

    aList[[key]] <- value
  }

  mydf <- data.frame(Results = c('A factor','Residuals', 'U factor', 'AxU', 'Residuals'),
                     Df = df,
                     Sum_Sq = ssq,
                     Mean_Sq = ms,
                     F_value = f,
                     P_value = p,
                     check.names = TRUE)
  grid.table(mydf)

  return(aList)
}
