#' Chi-Square (Pearson's method) Test statistic with privacy preserving
#'
#' Chi-Square Test statistic that preserves the privacy of the data sources. The statistic is performed for exactly two sample.
#'
#' @param params The <Parameters> that should be defined to compute the statistic.
#' @return A set of resulted components after the chi-square test execution, like p-value, df, X^2
#' @examples
#' We want to apply a chi-square test between the sex (M,F) and the age (old, young).
#' Firstly we should define the <Parameters> value:
#' chiGr1 <- GroupStat( list( c("sex", "=", "M") , c("sex", "=", "F") ))
#' chiGr2 <- GroupStat( list( c("age", "=", "old") , c("age", "=", "young")))
#' chiParam <- Parameters( list() , list(chiGr1, chiGr2))
#' chisq.test(chiParam)
#'
#' @export
chisq.test <- function(params)
{
  if(missing(params)) {
    stop("The argument 'params' should be filled in.")
  }

  if(class(params)[2] != "Parameters") {
    stop("The argument 'params' should be an instance of 'Parameters'.")
  }

  if(length(getGroups(params)) != 2) {
    stop("The chi square test statistic can be computed over exactly two sample group.")
  }

  for(gr in getGroups(params)) {
    tripList <- getTriples(gr)

    l <- rapply(tripList, function(x) head(x, 1))

    #if there are more than one duplicates
    if(length(unique(l)) != 1)
      stop("Each group should have the same variable at the set of triples.")
  }

  input <- .buildQuery(params)

  testObj <- .jnew("gr.demokritos.iit.radio.home.orchestration.Controller")
  result <- .jcall(testObj, "Ljava/util/HashMap;", "chiSquareParam", input)

  #remove later!!!
  #the results should be displayed in the end
  p1 <- proc.time()
  Sys.sleep(0.2)
  proc.time() - p1
  .convertChiSqResult(result)

  #print(result)

}

.buildWhereChiSq <- function(group, msg) {

  #where <- .jnew("gr.demokritos.iit.radio.home.entities.qfr.WhereClause")

  triples <- getTriples(group)

  for (triple in triples) {
    vect <- .jnew("java.util.Vector")

    for(val in triple) {
      .jrcall(vect, "add", val)
    }
    .jcall(msg, "V","addTriple", vect)

  }

  return(msg)
}

.convertChiSqResult <- function(result) {
  keySet<-.jrcall(result,"keySet")
  an_iter<-.jrcall(keySet,"iterator")
  aList <- list()

  while(.jrcall(an_iter,"hasNext")){
    key <- .jrcall(an_iter,"next");

    aList[[key]] <- .jrcall(result,"get",key)
  }

  return(aList)
}
