#' Simple Linear Regression statistic with privacy preserving
#'
#' Simple Linear Regression statistic that preserves the privacy of the data sources.
#' The statistic is performed for exactly one dependent and one independent variable.
#'
#' @param params The <Parameters> that should be defined to compute the statistic. Linear Regression needs two variables in the 'dependency' field of the object Parameters. The first string specifies the independent variable while the second the independent one.
#' @return A set of resulted components after the Linear Regression execution, like the coefficients (beta0 and beta1) and analytics (Sx, Sy, R2)
#' @examples
#' We want to apply a linear regression test between the 2 variables: value_time_before, value_time_after.
#' Firstly we should define the variables through the <Parameters>:
#' The first element in the Dependent list defines how much the variable is related linearly to the second one
#' lParam <- Parameters( list("value_time_before", "value_time_after") , NULL )
#' lr(lParam)
#' @export
lr <- function(params) {

  if(missing(params)) {
    stop("The argument 'params' should be filled in.")
  }

  if(class(params)[2] != "Parameters") {
    stop("The argument 'params' should be an instance of 'Parameters'.")
  }

  if(length(getDependency(params)) != 2) {
    stop("'lr' function needs exactly 1 dependent and one independent variables in the list of 'dependency'.")
  }

  input <- .buildQuery(params, cond = 'r')

  testObj <- .jnew("gr.demokritos.iit.radio.home.orchestration.Controller")
  result <- .jcall(testObj, "Ljava/util/HashMap;", "lrParam", input)

  #remove later!!!
  #the results should be displayed in the end
  p1 <- proc.time()
  Sys.sleep(0.2)
  proc.time() - p1

  r = .convertLrResult(result)
  .plotLine(r[["beta0"]], r[["beta1"]])

  return(r)

}

.convertLrResult <- function(result) {
  keySet<-.jrcall(result,"keySet")
  an_iter<-.jrcall(keySet,"iterator")
  aList <- list()

  while(.jrcall(an_iter,"hasNext")){
    key <- .jrcall(an_iter,"next");

    aList[[key]] <- .jrcall(result,"get",key)
  }

  return(aList)
}

#plot the regression line
.plotLine <- function(intercept, slope) {
  if(slope > 0.0)
    eq = paste0("y = ", round(intercept, 2), "*x + ", round(slope, 2))
  else
    eq = paste0("y = ", round(intercept, 2), "*x ", round(slope, 2))

  plot(c(-10,10), c(-10,10), type = "n", xlab="x", ylab="y", asp = 1, main=eq)

  abline(round(intercept, 2), round(slope, 2), col="blue")
}
