#' Normality with privacy preserving (D’ Agostino-Pearson omnibus method)
#'
#' Checkes whether a sample follows a normal distribution. It needs a single dependency attribute.
#'
#' @param params The <Parameters> that should be defined to compute whether a sample follows a normal distribution.  For this computation a single sample group is needed.
#' @param conf.level The confidence level of the interval. It is the opposite of significance level. Default: 0.95
#' @return A set of resulted components after the normality execution, like p-value, Zg1, Zg2 and if the sample is statistically significant normalised.
#' @examples
#' We want to check if the variable 'value_time_after' follows a normal distribution for the sample group where the people are female (sex=F).
#' Firstly we should define the <Parameters> with the DV and the IV:
#' nGr <- GroupStat( list( c("sex", "=", "F") ))
#' nParam <- Parameters( list("value_time_after") , list(nGr))
#' normality(nParam)
#' @export
normality <- function(params, conf.level=0.95)
{

  if(missing(params)) {
    stop("The argument 'params' should be filled in.")
  }

  if(class(params)[2] != "Parameters") {
    stop("The argument 'params' should be an instance of 'Parameters'.")
  }


  if(length(getGroups(params)) != 1) {
    stop("The normality test can be done only to one sample group.")
  }

  if(typeof(conf.level) != "double" || conf.level >= 1.0) {
    stop("You should specify the suitable formats in the arguments.\n
         - <conf.level>: double and below 1.0")
  }

  input <- .setInputNorm(params, conf.level)

  testObj <- .jnew("gr.demokritos.iit.radio.home.orchestration.Controller")
  result <- .jcall(testObj, "Ljava/util/HashMap;", "normalDistr", input)

  .convertNormalityResult(result)

  }

.setInputNorm <- function(params, conf.level) {
  # set the map with the fundamental information to compute the Normality test
  map <- .jnew("java.util.HashMap")

  .jrcall(map,"put", "conf.level", as.character(conf.level))

  input <- .buildQuery(params,map)

  return(input)

}

.convertNormalityResult <- function(result) {
  keySet<-.jrcall(result,"keySet")
  an_iter<-.jrcall(keySet,"iterator")
  aList <- list()

  while(.jrcall(an_iter,"hasNext")){
    key <- .jrcall(an_iter,"next");

    aList[[key]] <- .jrcall(result,"get",key)
  }

  return(aList)
}
